#include "UCalculator.h"
#include <math>
#include <locale>
#include <algorithm>
//---------------------------------------------------------------------------

// Comparison function used to sort strings by length - longest first.
bool CompareStringLength(const string& x, const string& y)
{
  return x.length() > y.length();
}
//---------------------------------------------------------------------------

void TStringParser::InsertDictionaryItem(const string& AToken)
{
  // Insert the string and resort the vector.
  if (!LocateDictionaryItem(AToken))
  {
    FDictionary.push_back(AToken);
    stable_sort(FDictionary.begin(), FDictionary.end(), CompareStringLength);
  }
}
//---------------------------------------------------------------------------

void TStringParser::DeleteDictionaryItem(const string& AToken)
{
  // Is it safe to erase end()?
  FDictionary.erase(find(FDictionary.begin(), FDictionary.end(), AToken));
}
//---------------------------------------------------------------------------

bool TStringParser::LocateDictionaryItem(const string& AToken) const
{
  return find(FDictionary.begin(), FDictionary.end(), AToken) != FDictionary.end();
}
//---------------------------------------------------------------------------

void TStringParser::ParseTheTargetString()
{
  // Get rid of the tokens from the previous target string.
  FTokens.clear();
  FTokenError = false;

  // Take a copy of the target string and trim out the white space.
  // I want to make this all a consistent case - not implemented.
  //string TokenString;
  //for (string::iterator I = FString.begin(); I != FString.end(); I++)
  //  if (*I != ' ') // Need to cater for other non-printing characters.
  //    TokenString.push_back(*I);

  // Is this more efficient? But it only caters for spaces - use remove_if().
  string TokenString = FString;
  TokenString.erase(std::remove(TokenString.begin(), TokenString.end(), ' '),
    TokenString.end());

  if (TokenString.empty())
    return;

  unsigned int Position = 0;
  while (true)
  {
    bool TokenFound;
    string NewToken;
    FindToken(TokenString, Position, TokenFound, NewToken);
    if (TokenFound)
    {
      FTokens.push_back(NewToken);
      Position += NewToken.length();
      if (Position == TokenString.length())
        break;
    }
    else
    {
      // Clear all the previously found tokens
      FTokens.clear();
      FTokenError = true;
      break;
    }
  }
}
//---------------------------------------------------------------------------

// Comparison function used to compare strings while ignoring case.
// Surely there is an easier way - but I'm new to this standard C++ library lark.
// Apparently this is correct...
bool CharsEqualIgnoreCase(char x, char y)
{
  static std::locale loc;
  return toupper(x, loc) == toupper(y, loc);
}
//---------------------------------------------------------------------------

void TStringParser::FindToken(const string& TokenString, int Position,
  bool& TokenFound, string& NewToken)
{
  TokenFound = true;
  vector<string>::iterator I = FDictionary.begin();
  string TestToken;
  while (I != FDictionary.end())
  {
    if ((*I).length() <= TokenString.length())
    {
      NewToken = *I;
      if (NewToken.length() != TestToken.length())
        TestToken = TokenString.substr(Position, NewToken.length());
      if (FCaseSensitive)
      {
        if (NewToken == TestToken)
          return;
      }
      else
      {
        if (std::equal(NewToken.begin(), NewToken.end(), TestToken.begin(), CharsEqualIgnoreCase))
          return;
      }
    }
    I++;
  }
  TokenFound = false;
  NewToken   = "";
}
//---------------------------------------------------------------------------

TExpressionEvaluator::TExpressionEvaluator()
{
  // Populate the lookup table of legal relationships between token types.
  SetupTokenRelationships();
  // Add the standard mathematical tokens which we allow...
  SetupDictionaryItems();

  FExpression = 0;
}
//---------------------------------------------------------------------------

TExpressionEvaluator::~TExpressionEvaluator()
{
  // Clear out any expression objects we may have left over - unlikely to be any.
  for (vector<TExpression*>::iterator I = FExpressions.begin(); I != FExpressions.end(); I++)
    delete *I;
  // Oops - this is deleted by the owning TFunction or during ProcessString
  // No wonder I was getting errors on destruction...
  //delete FExpression;
}
//---------------------------------------------------------------------------

void TExpressionEvaluator::SetupTokenRelationships()
{
  // I should make this table static or eomething. I'm sure there must be a better
  // way of doing this, but it works. It's just a bit clumsy - inelegant.

  // Test these pairs: ((, (), (sin, (*, (0, (e, (a, (f, (_
  FTokenRelationships[ttOpenParenthesis][ttOpenParenthesis]   = trAllowed;
  FTokenRelationships[ttOpenParenthesis][ttCloseParenthesis]  = trNotAllowed;
  FTokenRelationships[ttOpenParenthesis][ttUnary]             = trAllowed;
  FTokenRelationships[ttOpenParenthesis][ttBinary]            = trUnaryMinus;
  FTokenRelationships[ttOpenParenthesis][ttNumber]            = trAllowed;
  FTokenRelationships[ttOpenParenthesis][ttConstant]          = trAllowed;
  FTokenRelationships[ttOpenParenthesis][ttVariable]          = trAllowed;
  FTokenRelationships[ttOpenParenthesis][ttFunction]          = trAllowed;
  FTokenRelationships[ttOpenParenthesis][ttInvalidToken]      = trNotAllowed;

  // Test these pairs: )(, )), )sin, )*, )0, )e, )a, )f, )_
  FTokenRelationships[ttCloseParenthesis][ttOpenParenthesis]  = trImpliedProduct;
  FTokenRelationships[ttCloseParenthesis][ttCloseParenthesis] = trAllowed;
  FTokenRelationships[ttCloseParenthesis][ttUnary]            = trImpliedProduct;
  FTokenRelationships[ttCloseParenthesis][ttBinary]           = trAllowed;
  FTokenRelationships[ttCloseParenthesis][ttNumber]           = trImpliedProduct;
  FTokenRelationships[ttCloseParenthesis][ttConstant]         = trImpliedProduct;
  FTokenRelationships[ttCloseParenthesis][ttVariable]         = trImpliedProduct;
  FTokenRelationships[ttCloseParenthesis][ttFunction]         = trImpliedProduct;
  FTokenRelationships[ttCloseParenthesis][ttInvalidToken]     = trAllowed;

  // Test these pairs: sin(, sin), sinsin, sin*, sin0, sine, sina, sinf, sin_
  FTokenRelationships[ttUnary][ttOpenParenthesis]             = trAllowed;
  FTokenRelationships[ttUnary][ttCloseParenthesis]            = trNotAllowed;
  FTokenRelationships[ttUnary][ttUnary]                       = trAllowed;    //trNotAllowed; Changed to allow sintanx
  FTokenRelationships[ttUnary][ttBinary]                      = trUnaryMinus; //trNotAllowed; Changed to allow sin-3.
  FTokenRelationships[ttUnary][ttNumber]                      = trAllowed;
  FTokenRelationships[ttUnary][ttConstant]                    = trAllowed;
  FTokenRelationships[ttUnary][ttVariable]                    = trAllowed;
  FTokenRelationships[ttUnary][ttFunction]                    = trAllowed;
  FTokenRelationships[ttUnary][ttInvalidToken]                = trNotAllowed;

  // Test these pairs: *(, *), *sin, **, *0, *e, *a, *f, *_
  FTokenRelationships[ttBinary][ttOpenParenthesis]            = trAllowed;
  FTokenRelationships[ttBinary][ttCloseParenthesis]           = trNotAllowed;
  FTokenRelationships[ttBinary][ttUnary]                      = trAllowed;
  FTokenRelationships[ttBinary][ttBinary]                     = trUnaryMinus; //trNotAllowed; Changed to allow 3*-3.
  FTokenRelationships[ttBinary][ttNumber]                     = trAllowed;
  FTokenRelationships[ttBinary][ttConstant]                   = trAllowed;
  FTokenRelationships[ttBinary][ttVariable]                   = trAllowed;
  FTokenRelationships[ttBinary][ttFunction]                   = trAllowed;
  FTokenRelationships[ttBinary][ttInvalidToken]               = trNotAllowed;

  // Test these pairs: 0(, 0), 0sin, 0*, 00, 0e, 0a, 0f, 0_
  FTokenRelationships[ttNumber][ttOpenParenthesis]            = trImpliedProduct;
  FTokenRelationships[ttNumber][ttCloseParenthesis]           = trAllowed;
  FTokenRelationships[ttNumber][ttUnary]                      = trImpliedProduct;
  FTokenRelationships[ttNumber][ttBinary]                     = trAllowed;
  FTokenRelationships[ttNumber][ttNumber]                     = trImpliedProduct; // Will never happen though - parser concatenates.
  FTokenRelationships[ttNumber][ttConstant]                   = trImpliedProduct;
  FTokenRelationships[ttNumber][ttVariable]                   = trImpliedProduct;
  FTokenRelationships[ttNumber][ttFunction]                   = trImpliedProduct;
  FTokenRelationships[ttNumber][ttInvalidToken]               = trAllowed;

  // Test these pairs: e(, e), esin, e*, e0, ee ea, ef, e_
  FTokenRelationships[ttConstant][ttOpenParenthesis]          = trImpliedProduct;
  FTokenRelationships[ttConstant][ttCloseParenthesis]         = trAllowed;
  FTokenRelationships[ttConstant][ttUnary]                    = trImpliedProduct;
  FTokenRelationships[ttConstant][ttBinary]                   = trAllowed;
  FTokenRelationships[ttConstant][ttNumber]                   = trImpliedProduct;
  FTokenRelationships[ttConstant][ttConstant]                 = trImpliedProduct;
  FTokenRelationships[ttConstant][ttVariable]                 = trImpliedProduct;
  FTokenRelationships[ttConstant][ttFunction]                 = trImpliedProduct;
  FTokenRelationships[ttConstant][ttInvalidToken]             = trAllowed;

  // Test these pairs: a(, a), asin, a*, a0, ae, aa, af, a_
  FTokenRelationships[ttVariable][ttOpenParenthesis]          = trImpliedProduct;
  FTokenRelationships[ttVariable][ttCloseParenthesis]         = trAllowed;
  FTokenRelationships[ttVariable][ttUnary]                    = trImpliedProduct;
  FTokenRelationships[ttVariable][ttBinary]                   = trAllowed;
  FTokenRelationships[ttVariable][ttNumber]                   = trImpliedProduct;
  FTokenRelationships[ttVariable][ttConstant]                 = trImpliedProduct;
  FTokenRelationships[ttVariable][ttVariable]                 = trImpliedProduct;
  FTokenRelationships[ttVariable][ttFunction]                 = trImpliedProduct;
  FTokenRelationships[ttVariable][ttInvalidToken]             = trAllowed;

  // Test these pairs: f(, f), fsin, f*, f0, fe fa, ff, f_
  FTokenRelationships[ttFunction][ttOpenParenthesis]          = trImpliedProduct;
  FTokenRelationships[ttFunction][ttCloseParenthesis]         = trAllowed;
  FTokenRelationships[ttFunction][ttUnary]                    = trImpliedProduct;
  FTokenRelationships[ttFunction][ttBinary]                   = trAllowed;
  FTokenRelationships[ttFunction][ttNumber]                   = trImpliedProduct;
  FTokenRelationships[ttFunction][ttConstant]                 = trImpliedProduct;
  FTokenRelationships[ttFunction][ttVariable]                 = trImpliedProduct;
  FTokenRelationships[ttFunction][ttFunction]                 = trImpliedProduct;
  FTokenRelationships[ttFunction][ttInvalidToken]             = trAllowed;

  // Test these pairs: _(, _), _sin, _*, _0, _e _a, _f, __
  FTokenRelationships[ttInvalidToken][ttOpenParenthesis]      = trAllowed;
  FTokenRelationships[ttInvalidToken][ttCloseParenthesis]     = trNotAllowed;
  FTokenRelationships[ttInvalidToken][ttUnary]                = trAllowed;
  FTokenRelationships[ttInvalidToken][ttBinary]               = trUnaryMinus;
  FTokenRelationships[ttInvalidToken][ttNumber]               = trAllowed;
  FTokenRelationships[ttInvalidToken][ttConstant]             = trAllowed;
  FTokenRelationships[ttInvalidToken][ttVariable]             = trAllowed;
  FTokenRelationships[ttInvalidToken][ttFunction]             = trAllowed;
  FTokenRelationships[ttInvalidToken][ttInvalidToken]         = trAllowed;
}
//---------------------------------------------------------------------------

void TExpressionEvaluator::SetupDictionaryItems()
{
  // Parentheses have no evaluator since they merely modify priority.
  InsertDictionaryItem("(", ttOpenParenthesis,  stParenthesis, 0, 0, 0);
  InsertDictionaryItem(")", ttCloseParenthesis, stParenthesis, 0, 0, 0);

  // Numeric constants have no evaluatoe sonce they are returned directly.
  // Note that I make the parser treat numbers as a string of single characters.
  // It is for the next stage in the process to concatenate these into numeric
  // variables.
  InsertDictionaryItem("0", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("1", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("2", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("3", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("4", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("5", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("6", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("7", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("8", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem("9", ttNumber, stInteger,  1, GetNumericExpression, this);
  InsertDictionaryItem(".", ttNumber, stDecimal,  1, GetNumericExpression, this);
  InsertDictionaryItem("E", ttNumber, stExponent, 1, GetNumericExpression, this);

  // Key universal constant pi and e. Note that e^a is evaluated as "pow(e, a)"
  // whereas exp(a) is evaluated as "exp(a)" which may be more accurate.
  InsertDictionaryItem("pi", ttConstant, stPi, 1, GetConstantExpression, this);
  InsertDictionaryItem("e",  ttConstant, stE,  1, GetConstantExpression, this);

  // All the basic binary operators.
  InsertDictionaryItem("+", ttBinary, stAddition,    2, GetBinaryExpression, this);
  InsertDictionaryItem("-", ttBinary, stSubtraction, 2, GetBinaryExpression, this);
  InsertDictionaryItem("*", ttBinary, stProduct,     4, GetBinaryExpression, this);
  InsertDictionaryItem("/", ttBinary, stQuotient,    4, GetBinaryExpression, this);
  InsertDictionaryItem("^", ttBinary, stPower,       5, GetBinaryExpression, this);

  // Unary operators for trigonometric functions
  InsertDictionaryItem("sin",     ttUnary, stSin,     6, GetTrigonometricExpression, this);
  InsertDictionaryItem("cos",     ttUnary, stCos,     6, GetTrigonometricExpression, this);
  InsertDictionaryItem("tan",     ttUnary, stTan,     6, GetTrigonometricExpression, this);
  InsertDictionaryItem("arcsin",  ttUnary, stArcsin,  6, GetTrigonometricExpression, this);
  InsertDictionaryItem("arccos",  ttUnary, stArccos,  6, GetTrigonometricExpression, this);
  InsertDictionaryItem("arctan",  ttUnary, stArctan,  6, GetTrigonometricExpression, this);

  // Unary operators for hyperbolic functions
  InsertDictionaryItem("sinh",    ttUnary, stSinh,    6, GetHyperbolicExpression, this);
  InsertDictionaryItem("cosh",    ttUnary, stCosh,    6, GetHyperbolicExpression, this);
  InsertDictionaryItem("tanh",    ttUnary, stTanh,    6, GetHyperbolicExpression, this);
  InsertDictionaryItem("arcsinh", ttUnary, stArcsinh, 6, GetHyperbolicExpression, this);
  InsertDictionaryItem("arccosh", ttUnary, stArccosh, 6, GetHyperbolicExpression, this);
  InsertDictionaryItem("arctanh", ttUnary, stArctanh, 6, GetHyperbolicExpression, this);

  // Other unary operators
  InsertDictionaryItem("abs",     ttUnary, stAbs,     6, GetOtherUnaryExpression, this);
  InsertDictionaryItem("ln",      ttUnary, stLn,      6, GetOtherUnaryExpression, this);
  InsertDictionaryItem("log",     ttUnary, stLog,     6, GetOtherUnaryExpression, this);
  InsertDictionaryItem("exp",     ttUnary, stExp,     6, GetOtherUnaryExpression, this);
  InsertDictionaryItem("sqrt",    ttUnary, stSqrt,    6, GetOtherUnaryExpression, this);

  // One special case inserted on the fly.
  InsertPrivateDictionaryItem("-", "minus",  ttUnary, stMinus,  6, GetOtherUnaryExpression, this);

  // The client of this class may add its own tokens, for which it must
  // provide pointers to factory functions with the signature of
  // TExpressionFactory.
}
//---------------------------------------------------------------------------

void TExpressionEvaluator::InsertDictionaryItem(const string& Token, TTokenType Type,
  int Subtype, int Priority, TExpressionFactory Factory, void* Object)
{
  if (FDictionary.find(Token) == FDictionary.end())
  {
    TTokenInfo* TokenInfo = new TTokenInfo();
    TokenInfo->Token    = Token;
    TokenInfo->Private  = Token;
    TokenInfo->Type     = Type;
    TokenInfo->Subtype  = Subtype;
    TokenInfo->Priority = Priority;
    TokenInfo->Factory  = Factory;
    TokenInfo->Object   = Object;

    FDictionary.insert(pair<string, TTokenInfo*>(Token, TokenInfo));
    FParser.InsertDictionaryItem(Token);
  }
}
//---------------------------------------------------------------------------

void TExpressionEvaluator::InsertPrivateDictionaryItem(const string& Token, const string& Private,
  TTokenType Type, int Subtype, int Priority, TExpressionFactory Factory, void* Object)
{
  if (FDictionary.find(Private) == FDictionary.end())
  {
    TTokenInfo* TokenInfo = new TTokenInfo();
    TokenInfo->Token    = Token;
    TokenInfo->Private  = Private;
    TokenInfo->Type     = Type;
    TokenInfo->Subtype  = Subtype;
    TokenInfo->Priority = Priority;
    TokenInfo->Factory  = Factory;
    TokenInfo->Object   = Object;

    FDictionary.insert(pair<string, TTokenInfo*>(Private, TokenInfo));
    //FParser.InsertDictionaryItem(Token);
  }
}
//---------------------------------------------------------------------------

void TExpressionEvaluator::DeleteDictionaryItem(const string& AToken)
{
  map<string, TTokenInfo*>::iterator I = FDictionary.find(AToken);
  if (I != FDictionary.end())
  {
    FParser.DeleteDictionaryItem(AToken);
    delete I->second;
    FDictionary.erase(I);
  }
}
//---------------------------------------------------------------------------

const TTokenInfo* TExpressionEvaluator::LocateDictionaryItem(const string& AToken)
{
  map<string, TTokenInfo*>::iterator I = FDictionary.find(AToken);
  if (I != FDictionary.end())
    return I->second;
  return 0;
}
//---------------------------------------------------------------------------

const TTokenInfo* TExpressionEvaluator::LocatePrivateDictionaryItem(const string& AToken)
{
  map<string, TTokenInfo*>::iterator I = FDictionary.begin();
  while (I != FDictionary.end())
  {
    if (I->second->Private == AToken)
      return I->second;
  }
  return 0;
}
//---------------------------------------------------------------------------

void TExpressionEvaluator::ProcessString(const string& AString)
{
  // This class does not delete FExpression - that is the responsibility of
  // the client. The only exception is where the expression which has been
  // made is found to be invalid (below).
  FProcessedString = "";
  FExpression      = 0;
  FExpressions.clear();

  // Clear the list of normalised tokens.
  FTokens.clear();

  // Set the new string and parse it.
  FParser.ProcessString(AString);
  FTokenError = teNoError;
  if (FParser.GetTokenError())
    FTokenError = teIllegalToken;

  // Check whether any tokens at all were found.
  if (FTokenError == teNoError)
  {
    if (FParser.GetTokenCount() == 0)
      FTokenError = teNoTokens;
  }

  // Normalise the token list.
  if (FTokenError == teNoError)
    FTokenError = NormaliseTheTokenList();

  // Validate the expression sequence against a lookup table. Invalid
  // pairings cause a failure.
  if (FTokenError == teNoError)
    FTokenError = ValidateTheExpression();

  // Create the expression objects and pass them the tokens and parenthesis
  // level. Invalid tokens cause a failure. Delete all objects.
  if (FTokenError == teNoError)
    FTokenError = CreateExpressionObjects();

  // Run the evaluation procedure over the list of objects. Remove items as
  // they are made into a hierarchy - soon only one is left.
  if (FTokenError == teNoError)
    FTokenError = EvaluateTheExpression();

  if (FTokenError != teNoError)
  {
    // Delete any TExpression objects
    if (FExpressions.size() > 0)
    {
      for (vector<TExpression*>::iterator I = FExpressions.begin(); I != FExpressions.end(); I++)
        delete *I;
      FExpressions.clear();
    }
    // Just in case all the Expressions have been processed before the error.
    else if (FExpression)
      delete FExpression;

    FExpression = 0; // Probably superfluous
  }
}
//---------------------------------------------------------------------------

TTokenError TExpressionEvaluator::NormaliseTheTokenList()
{
  // This function simply builds a list of tokens which has been normalised
  // for maths purposes. TStringParser can only recognise a finite set of legal
  // tokens so it can't do numerics. My solution is to provide it with a set
  // of single character numerics and then concatenate contiguous series of these
  // characters into full numeric tokens. Still need to check their acceptabiltiy
  // though...

  bool IsNumber   = false;
  bool IsExponent = false;
  int  J = -1;
  for (int I = 0; I < FParser.GetTokenCount(); I++)
  {
    // Retrieving the info about a token.
    // It is safe to assume that all the tokens are legal at this point.
    // But the concatenated numbers will NOT generally be legal.

    const TTokenInfo* TokenInfo = LocateDictionaryItem(FParser.GetToken(I));

    bool Concatenated = false;
    // Concatenate adjacent number tokens.
    if (TokenInfo->Type == ttNumber && IsNumber)
    {
      FTokens[J] += TokenInfo->Private;
      Concatenated = true;
    }
    // Concatentate a "+" or "-" which follows "E".
    // This fudge is needed to allow scientific format numbers.
    else if (TokenInfo->Type == ttBinary && IsExponent &&
        (TokenInfo->Subtype == stAddition || TokenInfo->Subtype == stSubtraction))
    {
      FTokens[J] += TokenInfo->Private;
      Concatenated = true;
    }
    else
    {
      FTokens.push_back(TokenInfo->Private);
      J++;
    }

    // Prepare flags for the next loop
    IsNumber   = (TokenInfo->Type == ttNumber || Concatenated);
    IsExponent = (IsNumber && TokenInfo->Subtype == stExponent);

    if (IsNumber)
    {
      // If this is a number and the next item is not ttNumber
      // or there is no next item then test the number's acceptability
      if (I < FParser.GetTokenCount() - 1)
        TokenInfo = LocateDictionaryItem(FParser.GetToken(I + 1));

      // Note that the latter condition will lead to the mantissa of a
      // scientific format number being tested before the whole number.
      // Superfluous but not a problem.
      if ((I == FParser.GetTokenCount() - 1) || (TokenInfo->Type != ttNumber))
      {
        char* EndPtr;
#ifdef USE_LONG_DOUBLES
        _strtold(FTokens[J].c_str(), &EndPtr);
#else
        strtod(FTokens[J].c_str(), &EndPtr);
#endif
        // Better to write a function to test the format:
        // [ws][sn][ddd][.][ddd][fmt[sn]ddd] ?
        if (*EndPtr != '\0')
          return teInvalidNumber;
      }
    }
  }

  return teNoError;
}
//---------------------------------------------------------------------------

TTokenError TExpressionEvaluator::ValidateTheExpression()
{
  // Valid means that the sequence of tokens matches the rules we
  // have in the lookup table.
  TTokenType Type = ttInvalidToken;
  vector<string>::iterator I = FTokens.begin();
  while (I != FTokens.end())
  {
    //  Find the token type of the next token in the sequence.
    const TTokenInfo* TokenInfo = LocateDictionaryItem(*I);
    if (!TokenInfo)
      // Hardcoded catch-all for concatentated numbers.
      TokenInfo = LocateDictionaryItem("0");

    // Here we build up the case normalised version of the
    // original expression string. Numbers are a special case.
    // We show the public version of the token to the user.
    if (TokenInfo->Type != ttNumber)
      FProcessedString += TokenInfo->Token;
    else
    {
      // Convert the token into a normalised format.
      // Is this always desirable?
//#ifdef USE_LONG_DOUBLES
//      TDouble Value = _strtold(FTokens[I].c_str(), 0);
//#else
//      TDouble Value = strtod(FTokens[I].c_str(), 0);
//#endif
      FProcessedString += *I;
    }

    switch (IsLegalSequence(Type, TokenInfo->Type))
    {
      case trNotAllowed:
        return teBadSequence;
      case trImpliedProduct:
        I = FTokens.insert(I, string("*")); // Hard-coded.
        I++;
        break;
      case trUnaryMinus:
      {
        // Substitute the "-" with "minus" to indicate that this
        // token is acting as a unary operator. This is the private
        // format.
        if (TokenInfo->Subtype == stSubtraction) //"-"
          *I = "minus"; // Hard-coded.
        else
          return teBadSequence;
      }
    }

    Type = TokenInfo->Type;
    I++;
  }

  if (IsLegalSequence(Type, ttInvalidToken) == trNotAllowed)
    return teBadSequence;

  return teNoError;
}
//---------------------------------------------------------------------------

TTokenError TExpressionEvaluator::CreateExpressionObjects()
{
  int ParenthesisLevel = 0; // This is used to calculate the priority of each operator

  TExpression* SubExpression;
  for (vector<string>::iterator I = FTokens.begin(); I != FTokens.end(); I++)
  {
    // It is safe to assume that all the tokens are legal at this point,
    // except for the concatenated numbers.
    const TTokenInfo* TokenInfo = LocateDictionaryItem(*I);
    if (!TokenInfo)
      // Hardcoded catch-all for concatentated numbers.
      TokenInfo = LocateDictionaryItem("0");

    // I strip out the parentheses. They are only used to assess the priority
    // of operators.
    switch (TokenInfo->Type)
    {
      case ttOpenParenthesis:
        ParenthesisLevel++;
        break;
      case ttCloseParenthesis:
        ParenthesisLevel--;
        if (ParenthesisLevel < 0)
          return teBadParentheses;
        break;
      default:
        // We could also have an error with memory allocation.
        // Need to catch it and exit the function gracefully.
        SubExpression = TokenInfo->Factory(TokenInfo->Object, TokenInfo->Subtype);
        //SubExpression->FEvaluator   = this;
        SubExpression->FTokenInfo   = TokenInfo;
        SubExpression->FParentheses = ParenthesisLevel;
        if (TokenInfo->Type == ttNumber)
        {
          SubExpression->FNumeric = *I;
#ifdef USE_LONG_DOUBLES
          SubExpression->FValue = _strtold((*I).c_str(), 0);
#else
          SubExpression->FValue = strtod((*I).c_str(), 0);
#endif
        }
        FExpressions.push_back(SubExpression);
    }
  }

  if (ParenthesisLevel != 0)
    return teBadParentheses;
  else
    return teNoError;
}
//---------------------------------------------------------------------------

TExpression* TExpressionEvaluator::CreateExpressionObject(const string& AToken, int ParenthesisLevel)
{
  // This public function is used by clients to create new expression objects from
  // a string token. They don't need to know anything about the token's implementation
  // other than its string representation. The main reason for adding this
  // method is to allow TExpression objects to create their own partial derivatives.

  // Retrieving the info about a token.
  const TTokenInfo* TokenInfo = LocateDictionaryItem(AToken);
  if (!TokenInfo)
    // Hardcoded catch-all for concatentated numbers.
    TokenInfo = LocateDictionaryItem("0");

  // I strip out the parentheses. They are only used to assess the priority
  // of operators.
  if (TokenInfo->Type != ttOpenParenthesis &&
      TokenInfo->Type != ttCloseParenthesis)
  {
    // We could also have an error with memory allocation.
    // Need to catch it and exit the function gracefully.
    TExpression* Result = TokenInfo->Factory(TokenInfo->Object, TokenInfo->Subtype);
    //SubExpression->FEvaluator   = this;
    Result->FTokenInfo   = TokenInfo;
    Result->FParentheses = ParenthesisLevel;
    if (TokenInfo->Type == ttNumber)
    {
      Result->FNumeric = AToken;
#ifdef USE_LONG_DOUBLES
      Result->FValue = _strtold(AToken.c_str(), 0);
#else
      Result->FValue = strtod(AToken.c_str(), 0);
#endif
    }
    return Result;
  }

  return 0;
}
//---------------------------------------------------------------------------

int TExpressionEvaluator::GetHighestPriority(int Limit)
{
  // This simple function returns the highest priority operator
  // in FExpressions below a limiting priority. This doesn't seem
  // terribly effient but is only used while working out a new expression.
  // Note that we only care about binary and unary operators.

  int Result = 0;

  for (vector<TExpression*>::iterator I = FExpressions.begin(); I != FExpressions.end(); I++)
  {
    if ((*I)->GetPriority() > Result && ((*I)->GetPriority() < Limit || Limit == 0) &&
      ((*I)->GetTokenInfo()->Type == ttUnary || (*I)->GetTokenInfo()->Type == ttBinary))
      Result = (*I)->GetPriority();
  }

  return Result;
}
//---------------------------------------------------------------------------

TTokenError TExpressionEvaluator::EvaluateTheExpression()
{
  // Start from the highest priority and work down.
  // Start from the front of the expression and works forewards.
  // Work out the effect of operators on other items:
  //   a) a unary operator takes the following item out of the list, and
  //   b) a binary operator takes the precending and following items out
  //      of the list.
  // Eventually one item remains - this is the result expression we want.
  // Remove the last item and assign it to FExpression;

  while (FExpressions.size() > 1)
  {
    int Priority = GetHighestPriority(Priority);

    for (vector<TExpression*>::iterator I = FExpressions.begin(); I != FExpressions.end(); I++)
    {
      if ((*I)->GetPriority() == Priority)
      {
        if ((*I)->GetTokenInfo()->Type == ttUnary)
        {
          // Move the following item into the Expression since it must be the
          // operand. Then indicate to remove it from the list.
          (*I)->FOperand1 = *(I + 1);
          (*I)->FOperand1->FRemoved = true;
        }
        else // this must be ttBinary
        {
          // Move the preceding and following items into the Expression since
          // they must be the operands. Then remove them from the list. Also
          // need to decrement the position counter.
          (*I)->FOperand1 = *(I - 1);
          (*I)->FOperand2 = *(I + 1);
          (*I)->FOperand1->FRemoved = true;
          (*I)->FOperand2->FRemoved = true;
          FExpressions.erase(I + 1);
          FExpressions.erase(I - 1);  // Order is important here
          I--; // Adjust for item's the removal - not sure about this with the STL
        }
      }
    }

    // Now delete the items which were Removed from the list.
    for (vector<TExpression*>::iterator I = FExpressions.begin(); I != FExpressions.end(); I++)
    {
      if ((*I)->FRemoved)
      {
        I = FExpressions.erase(I);
        I--; // Adjust for item's the removal
      }
    }
  }

  if (FExpressions.size() == 1)
  {
    FExpression = *FExpressions.begin();
    FExpressions.clear();
  }

  return teNoError; // Placeholder
}
//---------------------------------------------------------------------------

TTokenResult TExpressionEvaluator::IsLegalSequence(TTokenType FirstToken,
    TTokenType SecondToken) const
{
  // This step of indirection allows for an alternative implementation later.
  return FTokenRelationships[FirstToken][SecondToken];
}
//---------------------------------------------------------------------------

TExpression* TExpressionEvaluator::GetConstantExpression(void* This, int Subtype)
{
  return new TConstantExpression(Subtype);
}
//---------------------------------------------------------------------------

TExpression* TExpressionEvaluator::GetNumericExpression(void* This, int Subtype)
{
  return new TNumericExpression();
}
//---------------------------------------------------------------------------

TExpression* TExpressionEvaluator::GetTrigonometricExpression(void* This, int Subtype)
{
  switch (Subtype)
  {
    case stSin:    return new TSinExpression();
    case stCos:    return new TCosExpression();
    case stTan:    return new TTanExpression();
    case stArcsin: return new TArcSinExpression();
    case stArccos: return new TArcCosExpression();
    case stArctan: return new TArcTanExpression();
  }
  return 0;
}
//---------------------------------------------------------------------------

TExpression* TExpressionEvaluator::GetHyperbolicExpression(void* This, int Subtype)
{
  switch (Subtype)
  {
    case stSinh:    return new TSinhExpression();
    case stCosh:    return new TCoshExpression();
    case stTanh:    return new TTanhExpression();
    case stArcsinh: return new TArcSinhExpression();
    case stArccosh: return new TArcCoshExpression();
    case stArctanh: return new TArcTanhExpression();
  }
  return 0;
}
//---------------------------------------------------------------------------

TExpression* TExpressionEvaluator::GetBinaryExpression(void* This, int Subtype)
{
  switch (Subtype)
  {
    case stAddition:    return new TAdditionExpression();
    case stSubtraction: return new TSubtractionExpression();
    case stProduct:     return new TProductExpression();
    case stQuotient:    return new TQuotientExpression();
    case stPower:       return new TPowerExpression();
  }
  return 0;
}
//---------------------------------------------------------------------------

TExpression* TExpressionEvaluator::GetOtherUnaryExpression(void* This, int Subtype)
{
  switch (Subtype)
  {
    case stAbs:   return new TAbsExpression();
    case stLn:    return new TLnExpression();
    case stLog:   return new TLogExpression();
    case stExp:   return new TExpExpression();
    case stMinus: return new TMinusExpression();
    case stSqrt:  return new TSqrtExpression();
  }
  return 0;
}
//---------------------------------------------------------------------------

TExpression::TExpression()
{
  FTokenInfo   = 0;
  FOperand1    = 0;
  FOperand2    = 0;
  FNumeric     = "";
  FParentheses = 0;
  FValue       = 0.0;
  FRemoved     = false;
}
//---------------------------------------------------------------------------

TExpression::~TExpression()
{
  // This frees up the allocated memory of each subexpression and cascades
  // until the whole expression has been de-allocated - VITAL STEP.
  if (FOperand1)
    delete FOperand1;
  if (FOperand2)
    delete FOperand2;
}
//---------------------------------------------------------------------------

void TExpression::IncrementParentheses()
{
  FParentheses++;
  if (FOperand1)
    FOperand1->IncrementParentheses();
  if (FOperand2)
    FOperand2->IncrementParentheses();
}
//---------------------------------------------------------------------------
void TExpression::DecrementParentheses()
{
  FParentheses--;
  if (FOperand1)
    FOperand1->DecrementParentheses();
  if (FOperand2)
    FOperand2->DecrementParentheses();
}
//---------------------------------------------------------------------------

/*
TExpression* TExpression::CreateDuplicateExpression()
{
  // Recurse the duplication into the first operand.
  TExpression* Operand1Duplicate = NULL;
  if (FOperand1)
    Operand1Duplicate = FOperand1->CreateDuplicateExpression();

  // Recurse the duplication into the second operand.
  TExpression* Operand2Duplicate = NULL;
  if (FOperand2)
    Operand2Duplicate = FOperand2->CreateDuplicateExpression();

  // Duplicate the current expression
  TExpression* Duplicate;
  if (TokenInfo->Type == ttNumber)
    Duplicate = Evaluator->CreateExpressionObject(Token, Parentheses); // For numeric expressions.
  else
    Duplicate = Evaluator->CreateExpressionObject(TokenInfo->Private, Parentheses);

  Duplicate->Operand1 = Operand1Duplicate;
  Duplicate->Operand2 = Operand2Duplicate;

  return Duplicate;
}
*/
//---------------------------------------------------------------------------

int TExpression::GetPriority() const
{
  return FParentheses * 10 + FTokenInfo->Priority;
}
//---------------------------------------------------------------------------

string TExpression::GetRPNString() const
{
  // Recurse down the tree to build up an RPN representation of the expression.
  switch (FTokenInfo->Type)
  {
    case ttNumber:
      return FNumeric;
    case ttUnary:
      if (GetIsConstant())
        return GetValueAsString();
      else
        return FOperand1->GetRPNString() + " " + FTokenInfo->Token;
    case ttBinary:
      if (GetIsConstant())
        return GetValueAsString();
      else
        return FOperand2->GetRPNString() + " " + FOperand1->GetRPNString() + " " + FTokenInfo->Token;
    //case ttVariable:
    //case ttFunction:
    //case ttConstant:
    default:
      return FTokenInfo->Token;
  }
}
//---------------------------------------------------------------------------

string TExpression::GetNaturalString() const
{
  // Recurse down the tree to build up a "natural" string representation of the expression.
  switch (FTokenInfo->Type)
  {
    case ttNumber:
      return FNumeric;
    case ttUnary:
      if (GetIsConstant())
        return GetValueAsString();
      else
        return FTokenInfo->Token + GetOperandString(FOperand1);
    case ttBinary:
      if (GetIsConstant())
        return GetValueAsString();
      else
        return GetOperandString(FOperand1) + FTokenInfo->Token + GetOperandString(FOperand2);
    //case ttVariable:
    //case ttFunction:
    //case ttConstant:
    default:
      return FTokenInfo->Token;
  }
}
//---------------------------------------------------------------------------

string TExpression::GetOperandString(const TExpression* Operand) const
{
  if (Operand->FParentheses > FParentheses)
    return "(" + Operand->GetNaturalString() + ")";
  else
    return Operand->GetNaturalString();
}
//---------------------------------------------------------------------------

TDouble TConstantExpression::GetValue() const
{
  switch (FConstantType)
  {
    case stPi: return TDouble(3.14159265358979323846264338327950288419716939);
    case stE:  return TDouble(2.71828182845904523536);
    // We can easily add more constants here as long as they are also
    // distinguished in FConstantType.
  }
  return TDouble(0);
}
//---------------------------------------------------------------------------

TDouble TNumericExpression::GetValue() const
{
  return TDouble(FValue);
}
//---------------------------------------------------------------------------

TDouble TMinusExpression::GetValue() const
{
  return TDouble(-(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TAdditionExpression::GetValue() const
{
  return TDouble((FOperand1->GetValue()) + (FOperand2->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TSubtractionExpression::GetValue() const
{
  return TDouble((FOperand1->GetValue()) - (FOperand2->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TProductExpression::GetValue() const
{
  return TDouble((FOperand1->GetValue()) * (FOperand2->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TQuotientExpression::GetValue() const
{
  return TDouble((FOperand1->GetValue()) / (FOperand2->GetValue()));
}
//---------------------------------------------------------------------------

#ifdef USE_LONG_DOUBLES

TDouble TAbsExpression::GetValue() const
{
  return TDouble(fabsl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TSinExpression::GetValue() const
{
  return TDouble(sinl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TCosExpression::GetValue() const
{
  return TDouble(cosl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TTanExpression::GetValue() const
{
  return TDouble(tanl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TArcSinExpression::GetValue() const
{
  return TDouble(asinl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TArcCosExpression::GetValue() const
{
  return acosl(FOperand1->GetValue());
}
//---------------------------------------------------------------------------

TDouble TArcTanExpression::GetValue() const
{
  return TDouble(atanl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TLnExpression::GetValue() const
{
  return TDouble(logl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TLogExpression::GetValue() const
{
  return TDouble(log10l(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TExpExpression::GetValue() const
{
  return TDouble(expl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TPowerExpression::GetValue() const
{
  return TDouble(powl(FOperand1->GetValue(), FOperand2->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TSinhExpression::GetValue() const
{
  return TDouble(sinhl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TCoshExpression::GetValue() const
{
  return TDouble(coshl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TTanhExpression::GetValue() const
{
  return TDouble(tanhl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TArcSinhExpression::GetValue() const
{
  TDouble Result(FOperand1->GetValue());
  return TDouble(logl(Result + sqrtl(powl(Result, 2) + 1)));
}
//---------------------------------------------------------------------------

TDouble TArcCoshExpression::GetValue() const
{
  TDouble Result(FOperand1->GetValue());
  return TDouble(logl(Result + sqrtl(powl(Result, 2) - 1)));
}
//---------------------------------------------------------------------------

TDouble TArcTanhExpression::GetValue() const
{
  TDouble Result(FOperand1->GetValue());
  return TDouble(0.5 * logl((1 + Result) / (1 - Result)));
}
//---------------------------------------------------------------------------

TDouble TSqrtExpression::GetValue() const
{
  return TDouble(sqrtl(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

#else

TDouble TAbsExpression::GetValue() const
{
  return TDouble(fabs(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TSinExpression::GetValue() const
{
  return TDouble(sin(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TCosExpression::GetValue() const
{
  return TDouble(cos(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TTanExpression::GetValue() const
{
  return TDouble(tan(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TArcSinExpression::GetValue() const
{
  return TDouble(asin(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TArcCosExpression::GetValue() const
{
  return TDouble(acos(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TArcTanExpression::GetValue() const
{
  return TDouble(atan(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TLnExpression::GetValue() const
{
  return TDouble(log(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TLogExpression::GetValue() const
{
  return TDouble(log10(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TExpExpression::GetValue() const
{
  return TDouble(exp(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TPowerExpression::GetValue() const
{
  return TDouble(pow(FOperand1->GetValue(), FOperand2->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TSinhExpression::GetValue() const
{
  return TDouble(sinh(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TCoshExpression::GetValue() const
{
  return TDouble(cosh(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TTanhExpression::GetValue() const
{
  return TDouble(tanh(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------

TDouble TArcSinhExpression::GetValue() const
{
  TDouble Result(FOperand1->GetValue());
  return TDouble(log(Result + sqrt(pow(Result, 2) + 1)));
}
//---------------------------------------------------------------------------

TDouble TArcCoshExpression::GetValue() const
{
  TDouble Result(FOperand1->GetValue());
  return TDouble(log(Result + sqrt(pow(Result, 2) - 1)));
}
//---------------------------------------------------------------------------

TDouble TArcTanhExpression::GetValue() const
{
  TDouble Result(FOperand1->GetValue());
  return TDouble(0.5 * log((1 + Result) / (1 - Result)));
}
//---------------------------------------------------------------------------

TDouble TSqrtExpression::GetValue() const
{
  return TDouble(sqrt(FOperand1->GetValue()));
}
//---------------------------------------------------------------------------
#endif

/*
TExpression* TConstantExpression::CreateDerivativeExpression(string AToken)
{
  // (constant)' = 0
  return NULL;
}
//---------------------------------------------------------------------------

TExpression* TNumericExpression::CreateDerivativeExpression(string AToken)
{
  // (number)' = 0
  return NULL;
}
//---------------------------------------------------------------------------

TExpression* TAbsExpression::CreateDerivativeExpression(string AToken)
{
  // (absF)' = F'*sgnF
  TExpression* Derivative;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // We don't support sgn yet so return F' for now.
  Derivative = Operand1Derivative;

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TSinExpression::CreateDerivativeExpression(string AToken)
{
  // (sinF)' = F'*cosF
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Cosine    = Evaluator->CreateExpressionObject("cos", Parentheses);
    Cosine->Operand1       = FOperand1->CreateDuplicateExpression();

    if (FOperand1->TokenInfo->Token == AToken)
    {
      Derivative = Cosine;
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product   = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1      = Operand1Derivative;
      Product->Operand2      = Cosine;

      Derivative = Product;
    }
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TCosExpression::CreateDerivativeExpression(string AToken)
{
  // (cosF)' = -F'*sinF
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Sine      = Evaluator->CreateExpressionObject("sin", Parentheses);
    Sine->Operand1         = FOperand1->CreateDuplicateExpression();

    TExpression* Result;
    if (FOperand1->TokenInfo->Token == AToken)
    {
      Result = Sine;
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product   = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1      = Operand1Derivative;
      Product->Operand2      = Sine;

      Result = Product;
    }

    TExpression* Minus     = Evaluator->CreateExpressionObject("minus", Parentheses);
    Minus->Operand1        = Result;

    Derivative = Minus;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TTanExpression::CreateDerivativeExpression(string AToken)
{
  // (tanF)' = F'/cosF^2
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Cosine    = Evaluator->CreateExpressionObject("cos", Parentheses);
    Cosine->Operand1       = FOperand1->CreateDuplicateExpression();

    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = Cosine;
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = Power;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TArcSinExpression::CreateDerivativeExpression(string AToken)
{
  // (arcsinF)' = F'/sqrt(1-F^2)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = FOperand1->CreateDuplicateExpression();
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Subtract  = Evaluator->CreateExpressionObject("-", Parentheses);
    Subtract->Operand1     = Evaluator->CreateExpressionObject("1", Parentheses);
    Subtract->Operand2     = Power;
    Subtract->IncrementParentheses();

    TExpression* Sqrt      = Evaluator->CreateExpressionObject("sqrt", Parentheses);
    Sqrt->Operand1         = Subtract;

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = Sqrt;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TArcCosExpression::CreateDerivativeExpression(string AToken)
{
  // (arccosF)' = -F'/sqrt(1-F^2)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = FOperand1->CreateDuplicateExpression();
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Subtract  = Evaluator->CreateExpressionObject("-", Parentheses);
    Subtract->Operand1     = Evaluator->CreateExpressionObject("1", Parentheses);
    Subtract->Operand2     = Power;
    Subtract->IncrementParentheses();

    TExpression* Sqrt      = Evaluator->CreateExpressionObject("sqrt", Parentheses);
    Sqrt->Operand1         = Subtract;

    TExpression* Minus     = Evaluator->CreateExpressionObject("minus", Parentheses);
    Minus->Operand1        = Operand1Derivative;

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Minus;
    Quotient->Operand2     = Sqrt;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TArcTanExpression::CreateDerivativeExpression(string AToken)
{
  // (arctanF)' = F'/(1+F^2)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = FOperand1->CreateDuplicateExpression();
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Addition  = Evaluator->CreateExpressionObject("+", Parentheses);
    Addition->Operand1     = Evaluator->CreateExpressionObject("1", Parentheses);
    Addition->Operand2     = Power;
    Addition->IncrementParentheses();

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = Addition;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TLnExpression::CreateDerivativeExpression(string AToken)
{
  // (lnF)' = F'/F
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = FOperand1->CreateDuplicateExpression();

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TLogExpression::CreateDerivativeExpression(string AToken)
{
  // (logF)' = F'/(F*ln10)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Log10     = Evaluator->CreateExpressionObject("ln", Parentheses);
    Log10->Operand1        = Evaluator->CreateExpressionObject("10", Parentheses);

    TExpression* Product   = Evaluator->CreateExpressionObject("*", Parentheses);
    Product->Operand1      = FOperand1->CreateDuplicateExpression();
    Product->Operand2      = Log10;
    Product->IncrementParentheses();

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = Product;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TExpExpression::CreateDerivativeExpression(string AToken)
{
  // (expF)' = F'*expF
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Exponential = Evaluator->CreateExpressionObject("exp", Parentheses);
    Exponential->Operand1    = FOperand1->CreateDuplicateExpression();

    if (FOperand1->TokenInfo->Token == AToken)
    {
      Derivative = Exponential;
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product   = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1      = Operand1Derivative;
      Product->Operand2      = Exponential;

      Derivative = Product;
    }
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TMinusExpression::CreateDerivativeExpression(string AToken)
{
  // (-F)' = -F'
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Minus = Evaluator->CreateExpressionObject("minus", Parentheses);
    Minus->Operand1    = Operand1Derivative;

    Derivative = Minus;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TAdditionExpression::CreateDerivativeExpression(string AToken)
{
  // (F+G)' = F'+G'
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Recurse the derivation into the first operand.
  TExpression* Operand2Derivative = NULL;
  if (FOperand2)
    Operand2Derivative = FOperand2->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative && Operand2Derivative)
  {
    TExpression* Addition = Evaluator->CreateExpressionObject("+", Parentheses);
    Addition->Operand1    = Operand1Derivative;
    Addition->Operand2    = Operand2Derivative;
    //Addition->IncrementParentheses();

    Derivative = Addition;
  }
  else if (Operand1Derivative)
  {
    Derivative = Operand1Derivative;
  }
  else if (Operand2Derivative)
  {
    Derivative = Operand2Derivative;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TSubtractionExpression::CreateDerivativeExpression(string AToken)
{
  // (F-G)' = F'-G'
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Recurse the derivation into the first operand.
  TExpression* Operand2Derivative = NULL;
  if (FOperand2)
    Operand2Derivative = FOperand2->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative && Operand2Derivative)
  {
    TExpression* Subtract = Evaluator->CreateExpressionObject("-", Parentheses);
    Subtract->Operand1    = Operand1Derivative;
    Subtract->Operand2    = Operand2Derivative;
    //Subtract->IncrementParentheses();

    Derivative = Subtract;
  }
  else if (Operand1Derivative)
  {
    Derivative = Operand1Derivative;
  }
  else if (Operand2Derivative)
  {
    TExpression* Minus = Evaluator->CreateExpressionObject("minus", Parentheses);
    Minus->Operand1    = Operand2Derivative;

    Derivative = Minus;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TProductExpression::CreateDerivativeExpression(string AToken)
{
  // (FG)' = F'*G+F*G'
  TExpression* Derivative = NULL;
  TExpression* LeftBit    = NULL;
  TExpression* RightBit   = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  if (Operand1Derivative)
  {
    if (FOperand1->TokenInfo->Token == AToken)
    {
      LeftBit              = FOperand2->CreateDuplicateExpression();
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1    = Operand1Derivative;
      Product->Operand2    = FOperand2->CreateDuplicateExpression();

      LeftBit = Product;
    }
  }

  // Recurse the derivation into the second operand.
  TExpression* Operand2Derivative = NULL;
  if (FOperand2)
    Operand2Derivative = FOperand2->CreateDerivativeExpression(AToken);

  if (Operand2Derivative)
  {
    if (FOperand2->TokenInfo->Token == AToken)
    {
      RightBit             = FOperand1->CreateDuplicateExpression();
      delete Operand2Derivative;
    }
    else
    {
      TExpression* Product = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1    = FOperand1->CreateDuplicateExpression();
      Product->Operand2    = Operand2Derivative;

      RightBit = Product;
    }
  }

  // Derive the current expression
  if (LeftBit && RightBit)
  {
    TExpression* Addition = Evaluator->CreateExpressionObject("+", Parentheses);
    Addition->Operand1    = LeftBit;
    Addition->Operand2    = RightBit;
    Addition->IncrementParentheses();

    Derivative = Addition;
  }
  else if (LeftBit)
  {
    Derivative = LeftBit;
  }
  else if (RightBit)
  {
    Derivative = RightBit;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TQuotientExpression::CreateDerivativeExpression(string AToken)
{
  // (F/G)' = (F'*G-F*G')/G^2 or F'/G or -F*G'/G^2 depending
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Recurse the derivation into the first operand.
  TExpression* Operand2Derivative = NULL;
  if (FOperand2)
    Operand2Derivative = FOperand2->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative && Operand2Derivative)
  {
    TExpression* LeftBit;
    if (FOperand1->TokenInfo->Token == AToken)
    {
      LeftBit             = FOperand2->CreateDuplicateExpression();
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product    = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1       = Operand1Derivative;
      Product->Operand2       = FOperand2->CreateDuplicateExpression();

      LeftBit = Product;
    }

    TExpression* RightBit;
    if (FOperand2->TokenInfo->Token == AToken)
    {
      RightBit              = FOperand1->CreateDuplicateExpression();
      delete Operand2Derivative;
    }
    else
    {
      TExpression* Product    = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1       = FOperand1->CreateDuplicateExpression();
      Product->Operand2       = Operand2Derivative;

      RightBit = Product;
    }

    TExpression* Subtract    = Evaluator->CreateExpressionObject("-", Parentheses);
    Subtract->Operand1       = LeftBit;
    Subtract->Operand2       = RightBit;
    Subtract->IncrementParentheses();

    TExpression* Power       = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1          = FOperand2->CreateDuplicateExpression();
    Power->Operand2          = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Quotient    = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1       = Subtract;
    Quotient->Operand2       = Power;

    Derivative = Quotient;
  }
  else if (Operand1Derivative)
  {
    TExpression* Quotient    = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1       = Operand1Derivative;
    Quotient->Operand2       = FOperand2->CreateDuplicateExpression();

    Derivative = Quotient;
  }
  else if (Operand2Derivative)
  {
    TExpression* RightBit;
    if (FOperand2->TokenInfo->Token == AToken)
    {
      RightBit                = FOperand1->CreateDuplicateExpression();
      delete Operand2Derivative;
    }
    else
    {
      TExpression* Product    = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1       = FOperand1->CreateDuplicateExpression();
      Product->Operand2       = Operand2Derivative;

      RightBit = Product;
    }

    TExpression* Minus       = Evaluator->CreateExpressionObject("minus", Parentheses);
    Minus->Operand1          = RightBit;

    TExpression* Power       = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1          = FOperand2->CreateDuplicateExpression();
    Power->Operand2          = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Quotient    = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1       = Minus;
    Quotient->Operand2       = Power;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TPowerExpression::CreateDerivativeExpression(string AToken)
{
  // (F^G)' = F'*G*F^(G-1)+lnF*G'*F^G
  TExpression* Derivative = NULL;
  TExpression* LeftBit    = NULL;
  TExpression* RightBit   = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  if (Operand1Derivative)
  {
    TExpression* Subtract    = Evaluator->CreateExpressionObject("-", Parentheses);
    Subtract->Operand1       = FOperand2->CreateDuplicateExpression();
    Subtract->Operand2       = Evaluator->CreateExpressionObject("1", Parentheses);
    Subtract->IncrementParentheses();

    TExpression* Power       = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1          = FOperand1->CreateDuplicateExpression();
    Power->Operand2          = Subtract;

    TExpression* Product1;
    if (FOperand1->TokenInfo->Token == AToken)
    {
      Product1             = FOperand2->CreateDuplicateExpression();
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product    = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1       = Operand1Derivative;
      Product->Operand2       = FOperand2->CreateDuplicateExpression();

      Product1 = Product;
    }

    TExpression* Product2    = Evaluator->CreateExpressionObject("*", Parentheses);
    Product2->Operand1       = Product1;
    Product2->Operand2       = Power;

    LeftBit = Product2;
  }

  // Recurse the derivation into the second operand.
  TExpression* Operand2Derivative = NULL;
  if (FOperand2)
    Operand2Derivative = FOperand2->CreateDerivativeExpression(AToken);

  if (Operand2Derivative)
  {
    TExpression* Natural     = Evaluator->CreateExpressionObject("ln", Parentheses);
    Natural->Operand1        = FOperand1->CreateDuplicateExpression();

    TExpression* Product1;
    if (FOperand2->TokenInfo->Token == AToken)
    {
      Product1               = Natural;
      delete Operand2Derivative;
    }
    else
    {
      TExpression* Product    = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1       = Natural;
      Product->Operand2       = Operand2Derivative;

      Product1 = Product;
    }

    TExpression* Product2    = Evaluator->CreateExpressionObject("*", Parentheses);
    Product2->Operand1       = Product1;
    Product2->Operand2       = CreateDuplicateExpression();

    RightBit = Product2;
  }

  // Derive the current expression
  if (LeftBit && RightBit)
  {
    TExpression* Addition = Evaluator->CreateExpressionObject("+", Parentheses);

    Addition->Operand1 = LeftBit;
    Addition->Operand2 = RightBit;

    Derivative = Addition;
  }
  else if (LeftBit)
  {
    Derivative = LeftBit;
  }
  else if (RightBit)
  {
    Derivative = RightBit;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TSinhExpression::CreateDerivativeExpression(string AToken)
{
  // (sinhF)' = F'*coshF
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Cosh      = Evaluator->CreateExpressionObject("cosh", Parentheses);
    Cosh->Operand1         = FOperand1->CreateDuplicateExpression();

    if (FOperand1->TokenInfo->Token == AToken)
    {
      Derivative = Cosh;
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product   = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1      = Operand1Derivative;
      Product->Operand2      = Cosh;

      Derivative = Product;
    }
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TCoshExpression::CreateDerivativeExpression(string AToken)
{
  // (coshF)' = F'*sinhF
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Sinh      = Evaluator->CreateExpressionObject("sinh", Parentheses);
    Sinh->Operand1         = FOperand1->CreateDuplicateExpression();

    if (FOperand1->TokenInfo->Token == AToken)
    {
      Derivative = Sinh;
      delete Operand1Derivative;
    }
    else
    {
      TExpression* Product   = Evaluator->CreateExpressionObject("*", Parentheses);
      Product->Operand1      = Operand1Derivative;
      Product->Operand2      = Sinh;

      Derivative = Product;
    }
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TTanhExpression::CreateDerivativeExpression(string AToken)
{
  // (tanhF)' = F'/coshF^2
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Cosh      = Evaluator->CreateExpressionObject("cosh", Parentheses);
    Cosh->Operand1         = FOperand1->CreateDuplicateExpression();

    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = Cosh;
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = Power;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TArcSinhExpression::CreateDerivativeExpression(string AToken)
{
  // (arcsinhF)' = F'/sqrt(F^2+1)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = FOperand1->CreateDuplicateExpression();
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Addition  = Evaluator->CreateExpressionObject("+", Parentheses);
    Addition->Operand1     = Power;
    Addition->Operand2     = Evaluator->CreateExpressionObject("1", Parentheses);
    Addition->IncrementParentheses();

    TExpression* Sqrt      = Evaluator->CreateExpressionObject("sqrt", Parentheses);
    Sqrt->Operand1         = Addition;

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = Sqrt;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TArcCoshExpression::CreateDerivativeExpression(string AToken)
{
  // (arccoshF)' = F'/sqrt(F^2-1)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = FOperand1->CreateDuplicateExpression();
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Subtract  = Evaluator->CreateExpressionObject("-", Parentheses);
    Subtract->Operand1     = Power;
    Subtract->Operand2     = Evaluator->CreateExpressionObject("1", Parentheses);
    Subtract->IncrementParentheses();

    TExpression* Sqrt      = Evaluator->CreateExpressionObject("sqrt", Parentheses);
    Sqrt->Operand1         = Subtract;

    TExpression* Quotient  = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1     = Operand1Derivative;
    Quotient->Operand2     = Sqrt;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TArcTanhExpression::CreateDerivativeExpression(string AToken)
{
  // (arctanhF)' = F'/(1-F^2)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Power     = Evaluator->CreateExpressionObject("^", Parentheses);
    Power->Operand1        = FOperand1->CreateDuplicateExpression();
    Power->Operand2        = Evaluator->CreateExpressionObject("2", Parentheses);

    TExpression* Subtract  = Evaluator->CreateExpressionObject("-", Parentheses);
    Subtract->Operand1     = Evaluator->CreateExpressionObject("1", Parentheses);
    Subtract->Operand2     = Power;
    Subtract->IncrementParentheses();

    TExpression* Quotient   = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1      = Operand1Derivative;
    Quotient->Operand2      = Subtract;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TSqrtExpression::CreateDerivativeExpression(string AToken)
{
  // (sqrtF)' = F'/(2*sqrtF)
  TExpression* Derivative = NULL;

  // Recurse the derivation into the first operand.
  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  // Derive the current expression
  if (Operand1Derivative)
  {
    TExpression* Sqrt     = Evaluator->CreateExpressionObject("sqrt", Parentheses);
    Sqrt->Operand1        = FOperand1->CreateDuplicateExpression();

    TExpression* Product  = Evaluator->CreateExpressionObject("*", Parentheses);
    Product->Operand1     = Evaluator->CreateExpressionObject("2", Parentheses);
    Product->Operand2     = Sqrt;
    Product->IncrementParentheses();

    TExpression* Quotient = Evaluator->CreateExpressionObject("/", Parentheses);
    Quotient->Operand1    = Operand1Derivative;
    Quotient->Operand2    = Product;

    Derivative = Quotient;
  }

  return Derivative;
}
//---------------------------------------------------------------------------
*/

#ifndef __STDC__
#ifdef USE_LONG_DOUBLES
int _matherrl(struct _exceptionl *Exception)
#else
int _matherr(struct _exception *Exception)
#endif
{
  throw EAStandardMathFunctionWentBang(Exception);
}
#endif // __STDC__
//---------------------------------------------------------------------------

TCalculator::TCalculator()
{
}
//---------------------------------------------------------------------------

TCalculator::~TCalculator()
{
  for (map<string, TFunction*>::iterator I = FFunctions.begin();
    I != FFunctions.end(); I++)
  {
    delete I->second;
  }

  for (map<string, TVariable*>::iterator I = FVariables.begin();
    I != FVariables.end(); I++)
  {
    delete I->second;
  }

  // For some reason the destructor is hanging here - this must be something
  // to do with the map template. This only seems to happen after I put in an
  // expression involving a variable and then change the variable value...
  // I am at a loss to explain this.
}
//---------------------------------------------------------------------------

void TCalculator::InsertVariable(const string& AToken)
{
  if (!FExpressionEvaluator.LocateDictionaryItem(AToken))
  {
    TVariable* AVariable = new TVariable(AToken, this);
    FExpressionEvaluator.InsertDictionaryItem(AToken, ttVariable, -1, 1, TVariable::GetVariableExpression, AVariable);
    FVariables.insert(pair<string, TVariable*>(AToken, AVariable));
  }
}
//---------------------------------------------------------------------------

void TCalculator::DeleteVariable(const string& AToken)
{
  map<string, TVariable*>::iterator I = FVariables.find(AToken);
  if (I != FVariables.end())
  {
    FExpressionEvaluator.DeleteDictionaryItem(AToken);
    delete I->second;
    FVariables.erase(I);
  }
}
//---------------------------------------------------------------------------

TVariable* TCalculator::GetVariable(const string& AToken)
{
  map<string, TVariable*>::iterator I = FVariables.find(AToken);
  if (I != FVariables.end())
    return I->second;
  else
    return 0;
}
//---------------------------------------------------------------------------

void TCalculator::InsertFunction(const string& AToken)
{
  if (!FExpressionEvaluator.LocateDictionaryItem(AToken))
  {
    TFunction* AFunction = new TFunction(AToken, this);
    FExpressionEvaluator.InsertDictionaryItem(AToken, ttFunction, -1, 1, TFunction::GetFunctionExpression, AFunction);
    FFunctions.insert(pair<string, TFunction*>(AToken, AFunction));
  }
}
//---------------------------------------------------------------------------

void TCalculator::DeleteFunction(const string& AToken)
{
  map<string, TFunction*>::iterator I = FFunctions.find(AToken);
  if (I != FFunctions.end())
  {
    FExpressionEvaluator.DeleteDictionaryItem(AToken);
    delete I->second;
    FFunctions.erase(I);
  }
}
//---------------------------------------------------------------------------

TFunction* TCalculator::GetFunction(const string& AToken)
{
  map<string, TFunction*>::iterator I = FFunctions.find(AToken);
  if (I != FFunctions.end())
    return I->second;
  else
    return 0;
}
//---------------------------------------------------------------------------

TTokenError TFunction::SetString(const string& AExpressionString)
{
  // Need to pass the string to (one of) TCalculator's TExpressionEvaluator(s).
  // Returns token error...

  // Delete the old function expression if there was one. Even if new function
  // fails - this is because the expression string is permanently erased in the
  // UI, perhaps a false assumption.
  if (FExpression)
  {
    delete FExpression;
    FExpression = 0;
  }

  // Do it!
  FCalculator->GetExpressionEvaluator().ProcessString(AExpressionString);

  TTokenError TokenError = FCalculator->GetExpressionEvaluator().GetTokenError();
  // Assign the resulting expression to the chosen function
  // Note: A blank expression yields ExpressionEvaluator.Expression == NULL.
  TExpression* AExpression = FCalculator->GetExpressionEvaluator().GetExpression();

  if (TokenError == teNoError && AExpression)
  {
    // We need to copy the TExpression details into a TFunctionExpression in
    // order to take advantage of the circular reference checking.
    FExpression = new TTopLevelExpression(AExpression);

    if (GetCircular())
      TokenError = teCircularReference;
  }

  if (TokenError == teNoError)
    FNormalisedString = FCalculator->GetExpressionEvaluator().GetProcessedString();
  else
    FNormalisedString = "";

  FTokenError = TokenError;
  return TokenError;
}
//---------------------------------------------------------------------------

// Static variable used to help with circular checking.
bool TTopLevelExpression::FCircularError;
//---------------------------------------------------------------------------

bool TTopLevelExpression::GetCircularError() const
{
  FCalledAlready = false;
  FCircularError = false;
  GetValue();
  return FCircularError;
}
//---------------------------------------------------------------------------

// Static variable used to help with exception handling.
bool TTopLevelExpression::FExceptionOccurred;
//---------------------------------------------------------------------------

TDouble TTopLevelExpression::GetValue() const
{
  TDouble Result = 0.0;

  // If a previous call to this function is on the
  // stack we have a circular reference.
  if (!FCalledAlready && !FCircularError)
  {
    // Indicate that this function call is on the stack.
    // This could be a reference counter which allowed recursion to a
    // predefined depth....
    FCalledAlready = true;

    try
    {
      // Actually retrieve the value.
      Result = FOperand1->GetValue();
    }
    catch(...)
    {
      FExceptionOccurred = true;
    }
  }
  else
    // This is a static member variable.
    FCircularError = true;

  // Indicate that this function call is off the stack.
  FCalledAlready = false;
  return Result;
}
//---------------------------------------------------------------------------

TDouble TFunctionExpression::GetValue() const
{
  if (*FFunction)
    return (*FFunction)->GetValue();
  return 0.0;
}
//---------------------------------------------------------------------------

bool TFunctionExpression::GetIsConstant() const
{
  if (*FFunction)
    return (*FFunction)->GetIsConstant();
  return 0.0;
}
//---------------------------------------------------------------------------

/*
TExpression* TVariableExpression::CreateDerivativeExpression(string AToken)
{
  // A variable's derivative is either unity or zero.
  TExpression* Derivative = NULL;

  if (TokenInfo->Type == ttVariable && AToken == TokenInfo->Token)
  {
    Derivative = Evaluator->CreateExpressionObject("1", Parentheses);
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TFunctionExpression::CreateDerivativeExpression(string AToken)
{
  // This is based on the idea that TFunction encapsulates a fixed number of partial
  // derivatives - first or second order only - which can be used for Dave's graphing.
  // A better solution might be to create a derivative in the ordinary way and
  // pass that back. This would be a lot like expanding out the dependencies before
  // taking the derivative - references to secondary functions would be lost, but
  // arbitrary depth derivatives would be possible on the result.
  TExpression* Derivative = NULL;

  if (*FFunction)
  {
    Derivative = (*FFunction)->CreateDerivativeExpression(AToken);
  }

  return Derivative;
}
//---------------------------------------------------------------------------

TExpression* TTopLevelExpression::CreateDerivativeExpression(const string& AToken)
{
  TExpression* Derivative = NULL;

  TExpression* Operand1Derivative = NULL;
  if (FOperand1)
    Operand1Derivative = FOperand1->CreateDerivativeExpression(AToken);

  if (Operand1Derivative)
  {
    //Derivative = new TTopLevelExpression();
    //Derivative->Operand1   = Operand1Derivative;


    Derivative = Operand1Derivative;
  }

  return Derivative;
}
*/
//---------------------------------------------------------------------------

TDouble TFunction::GetValue()
{
  if (FExpression)
  {
    // Reset the static variable property.
    FExpression->ResetExceptionOccurred();
    // Evaluate the expression.
    TDouble Result = FExpression->GetValue();

    if (FExpression->GetExceptionOccurred())
      FTokenError = teExceptionOccurred;
    return Result;
  }
  return 0.0;
}
//---------------------------------------------------------------------------


